/*
 * xgrk - Greek keyboard support for X
 *
 *  (c) 1999 Spiros Ioannou (sivann@softlab.ece.ntua.gr)
 *
 *	WindowMaker ver, Nicholas Christopoulos (wired_gr@yahoo.com)
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <signal.h>
#include <unistd.h>

#if defined(_WMAKER_VER)
#include <time.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>

#include <sys/wait.h>
#include <sys/param.h>
#include <sys/types.h>

#include <X11/extensions/shape.h>

#include "wmgeneral/wmgeneral.h"
#include "wmgeneral/misc.h"

#include "uk.xpm"
#include "gr.xpm"
#include "mask.xbm"

#else
#include "xgrk_uk.xpm"
#include "xgrk_el.xpm"
#endif

#include "xgrk.h"

#define POLL_INTERVAL 25000	/* microseconds */
#define RESET_DELAY   100000	/* microseconds */

char *KeyEventToString (XEvent * ev);
int map_init (Display *);
void setmap (Display *, int);
void descend_tree (Window, unsigned long);
int parseopts (int argc, char **argv);
void setXFlag (Window * , Pixmap * );

char version[]="1.1";

unsigned char cur_keymap[32], prev_keymap[32];

#if !defined(_WMAKER_VER)
Display *display;
#endif
char *xg_display = NULL;

KeySym *original_ks;
int min_keycode, max_keycode, numof_keycodes, syms_per_keycode;
unsigned char alt_l, meta_l,meta_r, alt_r, shift_l, shift_r, semicolon, f1;
unsigned char control_l,control_r;
int poll_mode = 1;
int flag = 1;
int verbose=0;
long int gx=0,gy=0;

KeySym *greek_plain_ks, *greek_accent_ks;
KeySym *greek_diaeresis_ks, *greek_diaeresis_accent_ks;


void
signal_hdl(int koko)
{
  fprintf(stderr,"\nxgrk:Reseting to UK...");
  setmap (display, -1);
  fprintf(stderr,"Done\n");
  exit (koko);
}

void
pipe_hdl(int koko)
{
  /* This sucks but it works*/
  display = XOpenDisplay (xg_display);
  fprintf(stderr,"xgrk:Reseting to UK...\n");
  setmap (display, -1);
  XCloseDisplay(display);
  fprintf(stderr,"Done\n");
  exit(koko);
}

void
alarm_hdl (int koko)
{
  descend_tree (DefaultRootWindow (display), KeyPressMask);
  signal (SIGALRM, alarm_hdl);
  alarm (2);
}

void
descend_tree (Window w, unsigned long type)
{
  int ret, i;
  static long int level = 0;
  unsigned int nchildren;
  Window root, parent, *children;

  alarm (0);
  /* printf("Descending tree level:%ld\n",level); */

  ret = XQueryTree (display, w, &root, &parent, &children, &nchildren);

  if (ret == False) {
    fprintf (stdout, "xgrk:Query error level:%ld\n", level);
    return;
  }
  if (!nchildren)
    return;

  if (w == root) {
    level = 0;
    XSelectInput (display, root, type);
  }
  level++;

  for (i = 0; i < nchildren; i++) {
    XSelectInput (display, children[i], type);
    descend_tree (children[i], type);
  }

  if (children)
    XFree (children);
}

#define IS_DOWN(keymap,keycode)  (int)((keymap[keycode/8]>>(keycode%8))&1)

int
main (int argc, char *argv[])
{
  char *string;
  int state = -1;
  int keycode = 0, kc;
  int prevkeycode = 0;
  int status;

  XEvent xev;
  Window root;

#if !defined(_WMAKER_VER)
  int screen, black;
  int width, height;
  unsigned long int root_height,root_width;
  Window win, win2;
  Pixmap pixmap_uk, pixmap_el;
  XpmAttributes attributes;
  XSizeHints wmhints;

  char **koko;
#endif


#if defined(_WMAKER_VER)
	// create the window
	openXwindow(argc, argv, uk, mask_bits, 64, 64);
	AddMouseRegion(8, 8, 48, 48, 60);
	RedrawWindow();
#endif

  parseopts (argc, argv);
  
#if !defined(_WMAKER_VER)
  display = XOpenDisplay (xg_display);
  if (display == NULL) {
    perror ("XOpenDisplay");
    exit (10);
  }
#endif

  XDisplayKeycodes (display, &min_keycode, &max_keycode);
  numof_keycodes = max_keycode - min_keycode + 1;
  original_ks = XGetKeyboardMapping (display, min_keycode,
				     numof_keycodes, &syms_per_keycode);
  map_init (display);

  root = DefaultRootWindow (display);

#if !defined(_WMAKER_VER)
  screen = DefaultScreen (display);
  black = BlackPixel (display, screen);
  root_height=DisplayHeight(display,screen);
  root_width=DisplayWidth(display,screen);
#endif

  signal(SIGHUP,signal_hdl);	
  signal(SIGINT,signal_hdl);	
  signal(SIGTERM,signal_hdl);	

  if (flag) {
    signal(SIGPIPE,pipe_hdl);	

#if !defined(_WMAKER_VER)
	//
	//	standard X - initilization
	//
    attributes.valuemask = 0;
    status = XpmCreatePixmapFromData (display, root, gb, &pixmap_uk,
				  NULL, &attributes);

    if (status != XpmSuccess) {
      fprintf (stderr, "%s\n",XpmGetErrorString (status));
      exit (-1);
    }

    status = XpmCreatePixmapFromData (display, root, greece, &pixmap_el,
				  NULL, NULL);

    if (status != XpmSuccess) {
      fprintf (stderr, "%s\n",XpmGetErrorString (status));
      exit (-2);
    }

    width = attributes.width;
    height = attributes.height;
    if (gx<0) gx+=root_width-gx-width;
    if (gy<0) gy=root_height-gy-height;

    win = XCreateSimpleWindow (display, root, gx,gy, width, height, 0,
			       black, black);
    win2 = XCreateSimpleWindow (display, win, 0, 0, width, height, 0,
				black, black);
    XStoreName (display, win, "xgrk");
   
    wmhints.flags=USSize|USPosition;
    wmhints.width=wmhints.min_width=wmhints.max_width=width;
    wmhints.height=wmhints.min_height=wmhints.max_height=height;
    XSetWMNormalHints(display,win,&wmhints);

    XSelectInput (display, win2, ButtonPressMask);

    XMapWindow (display, win);
    XMapWindow (display, win2);
    XSync(display,False);
#endif
  }

  if (flag)
#if defined(_WMAKER_VER)
	replaceXPM(uk);
#else
    setXFlag (&win2,&pixmap_uk);
#endif
  

	if (!poll_mode)	{
    	signal (SIGALRM, alarm_hdl);
    	descend_tree (root, KeyPressMask);
    	XSelectInput (display, root, SubstructureNotifyMask);
  		}

	if (poll_mode)
		XQueryKeymap (display, prev_keymap);

	// event loop 
	while ( 1 ) {

		if ( !poll_mode ) {
	    	XSync (display, True);		/* Alt repeats forever in XFree without this */
			XNextEvent(display, &xev);

#if defined(_WMAKER_VER)
				if	( xev.type == Expose )	{
					RedrawWindow();
					continue;
					}
#endif
	      	if (verbose)
				printf ("\nEvenT:%d\n", xev.type);

	      	keycode = xev.xkey.keycode;
	      	usleep (10000);
	    	}

	    else {
	      	if ( XPending(display) )	{
				XNextEvent(display, &xev);
				XSync(display,False);

#if defined(_WMAKER_VER)
				if	( xev.type == Expose )	{
					RedrawWindow();
					continue;
					}
#endif

				if (verbose)
					printf ("\nEvent2:%d\n", xev.type);
	      		}

			usleep (POLL_INTERVAL);	/* Poll interval */
	      	XQueryKeymap (display, cur_keymap);

	      	if (memcmp (cur_keymap, prev_keymap, 32) == 0)	{
				if (!flag && xev.type != ButtonPress) 
		  			continue;
				}

			keycode = -1;
			for (kc = min_keycode; kc <= max_keycode; kc++)	{
				if (IS_DOWN (cur_keymap, kc) - IS_DOWN (prev_keymap, kc) == 1) 
	  				keycode = kc;
				else if (IS_DOWN (cur_keymap, kc) - IS_DOWN (prev_keymap, kc) == -1) 
					keycode = -kc;
				}

			memcpy (prev_keymap, cur_keymap, 32);
    		} // else - !poll

	//
	// We have an event
	//
	if ( keycode == control_l ) {
    	if (state>=0) { 
			state=-1;
			setmap (display, -1);
			if (flag) {
#if defined(_WMAKER_VER)
				replaceXPM(uk);
#else
			    setXFlag (&win2,&pixmap_uk);
				XStoreName (display, win, "xgrk-English");
#endif
				}
      		}
    	}
  
	if ( poll_mode && keycode < 0 )	{
    	if (!flag || (flag && xev.type != ButtonPress))
			continue;
		}

    if (!poll_mode && (xev.type == MapNotify)) {
      usleep (10000);
      descend_tree (root, KeyPressMask);
      alarm (1);		/* Re - descent the tree after a while */
    }
    else if (poll_mode || xev.type == KeyPress || xev.type==ButtonPress) {
      if (!poll_mode) {
	string = KeyEventToString (&xev);
	if (string == NULL)
	  continue;
	if (verbose && strlen (string) == 1)
	  printf ("%c", *string);
      }

      if (verbose && !poll_mode)
	printf ("<%s>", string);
      if (verbose)
        printf ("keycode:%d\n", keycode);

      /*Alt-Shift , Meta-Shift */
      if (( (flag && xev.type == ButtonPress) || 
	    (keycode == shift_l &&
	  (prevkeycode == alt_l || prevkeycode == meta_l))  ||
	    (keycode == shift_r && 
	  (prevkeycode == alt_r || prevkeycode == meta_r)) ) /*&& state < 0*/) {
	if (state<0){
	  state = 0;
	  setmap (display, 0);
	  if (flag) {
#if defined(_WMAKER_VER)
		replaceXPM(gr);
#else
	    setXFlag (&win2,&pixmap_el);
	    XStoreName (display, win, "xgrk-Greek");
#endif
	  }
	}
	else
	{
	  state = -1;
	  setmap (display, -1);
	  if (flag){  
#if defined(_WMAKER_VER)
		replaceXPM(uk);
#else
	    setXFlag (&win2,&pixmap_uk);
		XStoreName (display, win, "xgrk-English");
#endif
	  }
	}
      }

      if (keycode == semicolon && state == 0 && 
	  ((prevkeycode != shift_r) && (prevkeycode != shift_l)) ) {
	setmap (display, 1);
	state = 1;
      }
      else if (keycode == semicolon && state == 0 && 
	  ((prevkeycode == shift_r) || (prevkeycode == shift_l)) ) {
	setmap (display, 2);
	state = 2;
      }
      else if (keycode == semicolon && state == 1 && 
	  ((prevkeycode == shift_r) || (prevkeycode == shift_l)) ) {
	setmap (display, 3);
	state = 3;
      }
      else if (state>0&&keycode  && keycode != shift_r && keycode != shift_l) {
	/* Race condition. Will the Xclient get the accented letter before 
	 * the map gets reset again? yes if we are slow enough... 
	 */
	usleep (RESET_DELAY);
	XSync (display, False);
	setmap (display, 0);
	state = 0;
      }

      fflush (stdout);
      prevkeycode = keycode;
    }
  }
}


#define KEY_BUFF_SIZE 256
static char key_buff[KEY_BUFF_SIZE];

char *
KeyEventToString (XEvent * ev)
{
  int count;
  char *tmp;
  KeySym ks;

  ks = 0;

  if (ev->type == KeyPress) {
    count = XLookupString ((XKeyEvent *) ev, key_buff, KEY_BUFF_SIZE, &ks, NULL);
    key_buff[count] = '\0';

    if (count == 0) {
      tmp = XKeysymToString (ks);
      if (tmp)
	strcpy (key_buff, tmp);
      else
	key_buff[0] = 0;
    }

    return key_buff;
  }
  else
    return NULL;
}



int
map_init (Display * d)
{

  int i;
  int keycode_size;
  unsigned char keycode;

  char letters[] =
  {'a', 'b', 'g', 'd', 'e', 'z', 'h', 'u', 'i', 'k', 'l', 'm', 'n', 'j',
   'o', 'p', 'r', 'w', 's', 't', 'y', 'f', 'x', 'c', 'v', 'q', ';'
  };

  /* Find the keycodes from the letters */

  for (i = 0; i < sizeof (letters); i++) {
    keycode = XKeysymToKeycode (display, letters[i]);
    greek_plain[i][0] = keycode;
    greek_accent[i][0] = keycode;
    greek_diaeresis[i][0] = keycode;
    greek_diaeresis_accent[i][0] = keycode;
  }
  f1 = XKeysymToKeycode (display, XStringToKeysym ("F1"));
  greek_plain[i][0] = f1;
  greek_accent[i][0] = f1;
  greek_diaeresis[i][0] = f1;
  greek_diaeresis_accent[i][0] = f1;

  alt_l = XKeysymToKeycode (display, XStringToKeysym ("Alt_L"));
  alt_r = XKeysymToKeycode (display, XStringToKeysym ("Alt_R"));
  meta_l = XKeysymToKeycode (display, XStringToKeysym ("Meta_L"));
  meta_r = XKeysymToKeycode (display, XStringToKeysym ("Meta_R"));
  shift_l = XKeysymToKeycode (display, XStringToKeysym ("Shift_L"));
  shift_r = XKeysymToKeycode (display, XStringToKeysym ("Shift_R"));
  semicolon = XKeysymToKeycode (display, XStringToKeysym ("semicolon"));
  control_l = XKeysymToKeycode (display, XStringToKeysym ("Control_L"));
  control_r = XKeysymToKeycode (display, XStringToKeysym ("Control_R"));
  
  if (verbose)
    printf("semicolon:%d\n",semicolon);

  /* Now, make the keysym tables */

  keycode_size = numof_keycodes * syms_per_keycode * sizeof (KeySym);
  greek_plain_ks = (KeySym *) malloc (keycode_size);
  greek_accent_ks = (KeySym *) malloc (keycode_size);
  greek_diaeresis_ks = (KeySym *) malloc (keycode_size);
  greek_diaeresis_accent_ks = (KeySym *) malloc (keycode_size);

  if (!(greek_plain_ks && greek_accent_ks && greek_diaeresis_ks &&
	greek_diaeresis_accent_ks)) {
    perror ("malloc of keysym tables");
    exit (errno);
  }

  memcpy (greek_plain_ks, original_ks, keycode_size);
  memcpy (greek_accent_ks, original_ks, keycode_size);
  memcpy (greek_diaeresis_ks, original_ks, keycode_size);
  memcpy (greek_diaeresis_accent_ks, original_ks, keycode_size);



  for (i = 0; i < 28; i++) {
    keycode = greek_plain[i][0];	/* keycodes are the same in all tables*/

    greek_plain_ks[(keycode - min_keycode) * syms_per_keycode] =
	greek_plain[i][1];
    greek_plain_ks[(keycode - min_keycode) * syms_per_keycode + 1] =
	greek_plain[i][2];

    greek_accent_ks[(keycode - min_keycode) * syms_per_keycode] =
	greek_accent[i][1];
    greek_accent_ks[(keycode - min_keycode) * syms_per_keycode + 1] =
	greek_accent[i][2];

    greek_diaeresis_ks[(keycode - min_keycode) * syms_per_keycode] =
	greek_diaeresis[i][1];
    greek_diaeresis_ks[(keycode - min_keycode) * syms_per_keycode + 1] =
	greek_diaeresis[i][2];

    greek_diaeresis_accent_ks[(keycode - min_keycode) * syms_per_keycode] =
	greek_diaeresis_accent[i][1];
    greek_diaeresis_accent_ks[(keycode - min_keycode) * syms_per_keycode + 1] =
	greek_diaeresis_accent[i][2];

  }

  return 1;
}

void
setmap (Display * d, int map_id)
{
  verbose && printf ("Switching to map %d\n", map_id);
  switch (map_id) {
  case 0:
    XChangeKeyboardMapping (display, min_keycode, syms_per_keycode,
			    greek_plain_ks, numof_keycodes);
    break;
  case 1:
    XChangeKeyboardMapping (display, min_keycode, syms_per_keycode,
			    greek_accent_ks, numof_keycodes);
    break;
  case 2:
    XChangeKeyboardMapping (display, min_keycode, syms_per_keycode,
			    greek_diaeresis_ks, numof_keycodes);
    break;
  case 3:
    XChangeKeyboardMapping (display, min_keycode, syms_per_keycode,
			    greek_diaeresis_accent_ks, numof_keycodes);
    break;
  default:
    XChangeKeyboardMapping (display, min_keycode, syms_per_keycode,
			    original_ks, numof_keycodes);
  }

  XSync (display, False);
}

int
parseopts (int argc, char **argv)
{
  int i;


  for (i = 1; i < argc; i++) {
    if (strcmp (argv[i], "-poll") == 0) {
      poll_mode = 1;
    }
    else if (strcmp (argv[i], "-event") == 0) {
      poll_mode = 0;
    }
    else if (strcmp (argv[i], "-v") == 0) {
      verbose = 1;
    }
    else if (strcmp (argv[i], "-noflag") == 0) {
      flag = 0;
    }
    else if (strcmp (argv[i], "-geometry") == 0) {
      if (argc<i+2) {
	fprintf (stderr, "xgrk: -geometry: parse error\n");
	exit (0);
      }
      i++;
      if (sscanf(argv[i],"%ld%ld",&gx,&gy) < 2) {
	printf("xgrk: -geometry: not enough parameters\n");
	exit(0);
      }
      printf("gx:%ld,gy:%ld",gx,gy);
    }
    else if (strcmp (argv[i], "-display") == 0) {
      if (argc<=i+1) {
	fprintf (stderr, "xgrk: -display: parse error\n");
	exit (0);
      }
      i++;
      xg_display = argv[i];
    }
    else {
      fprintf(stderr, "xgrk: invalid argument %s\n\n", argv[i]);
      fprintf(stderr, "xgrk Version %s (c) Spiros Ioannou 1999\n",version);
      fprintf(stderr, 
	  "Usage: xgrk [-poll] [-event] [-v] [-noflag] [-display display]"
	  "[-geometry x y]\n");
      fprintf(stderr,
	  "\n-poll\t\tpoll the xserver for keyboard stat every %d us (default)\n",
	  POLL_INTERVAL);
      fprintf(stderr,"-event\t\tevent based keyboard monitoring\n");
      fprintf(stderr,"-noflag\t\tdon't display a window\n");
      fprintf(stderr,"-display\tdisplay to connect to\n");
      fprintf(stderr,"-geometry\twhere to popup\n");
      fprintf(stderr,"-v\t\tbe verbose \n\n");
      exit (0);
    }
  }
  return i;
}

void setXFlag (Window * win, Pixmap * pixmap)
{
  XSetWindowBackgroundPixmap (display, *win, *pixmap);
  XUnmapWindow (display,* win);
  XMapWindow (display,* win);
}
